from typing import Literal, Union


def format_messages(messages: Union[str, list[dict]] = None) -> list[dict]:
    try:
        if isinstance(messages, str):
            return [{"role": "user", "content": messages}]
        elif isinstance(messages, list) and all(isinstance(m, dict) for m in messages):
            return messages
        else:
            raise ValueError("Invalid message format")
    except Exception as e:
        print(f"Error formatting messages: {e}")
        return []


def does_system_prompt_exist(messages: list[dict]) -> bool:
    return any(message.get("role") == "system" for message in messages)


def swap_system_prompt(
    system_prompt: dict = None, messages: Union[str, list[dict[str, str]]] = None
):
    messages = format_messages(messages)

    for message in messages:
        # Check if a system message exists
        if message.get("role") == "system":
            # If a system message exists, swap it with the system_prompt
            message = system_prompt
            # Move the system_prompt to the beginning of the list
            messages.insert(0, message)
            # Remove the system_prompt from its original position
            messages.remove(message)
            break

        else:
            # If no system message exists, add the system_prompt to the beginning of the list
            messages.insert(0, system_prompt)
            break

    # Remove any duplicate system messages
    while len([message for message in messages if message.get("role") == "system"]) > 1:
        messages.pop()

    return messages



__all__ = [
    "Agents",
    "completion",
    "embedding",
    "image",
    "llm",
    "logger",
    "Memory",
    "speak",
    "tools",
    "terminal",
    "transcribe",
    "classify",
    "code",
    "create_system_prompt",
    "extract",
    "function",
    "generate",
    "least_to_most",
    "optimize_system_prompt",
    "plan",
    "self_consistency",
    "self_refine",
    "step_back",
    "tree_of_thought",
]

from ._client.utils.loader import Loader
from ._client import tools
from ._client.llm import (
    classify,
    code,
    completion,
    create_system_prompt,
    extract,
    function,
    generate,
    least_to_most,
    optimize_system_prompt,
    plan,
    self_consistency,
    self_refine,
    step_back,
    tree_of_thought,
)


class Agents(Loader):
    pass


Agents.init("zyx._client.agents", "Agents")


class completion(Loader):
    pass


completion.init("zyx._client.completion", "completion")


class embedding(Loader):
    pass


embedding.init("litellm.main", "embedding")


class image(Loader):
    pass


image.init("zyx._client.multimodal", "image")


class logger(Loader):
    pass


logger.init("loguru", "logger")


class Memory(Loader):
    pass


Memory.init("zyx._client.memory", "Memory")


class speak(Loader):
    pass


speak.init("zyx._client.multimodal", "speak")


class terminal(Loader):
    pass


terminal.init("zyx._client.terminal", "terminal")


class transcribe(Loader):
    pass


transcribe.init("zyx._client.multimodal", "transcribe")



__all__ = [
    "Client",
    "classify",
    "code",
    "completion",
    "create_system_prompt",
    "extract",
    "function",
    "generate",
    "least_to_most",
    "optimize_system_prompt",
    "plan",
    "self_consistency",
    "self_refine",
    "step_back",
    "tree_of_thought",
]


from ..utils.loader import Loader


class Client(Loader):
    pass


Client.init("zyx._client.completion", "CompletionClient")


class classify(Loader):
    pass


classify.init("zyx._client.llm.classify", "classify")


class completion(Loader):
    pass


completion.init("zyx._client.completion", "completion")


class code(Loader):
    pass


code.init("zyx._client.llm.code", "code")


class create_system_prompt(Loader):
    pass


create_system_prompt.init(
    "zyx._client.llm.create_system_prompt", "create_system_prompt"
)


class extract(Loader):
    pass


extract.init("zyx._client.llm.extract", "extract")


class function(Loader):
    pass


function.init("zyx._client.llm.function", "function")


class generate(Loader):
    pass


generate.init("zyx._client.llm.generate", "generate")


class least_to_most(Loader):
    pass


least_to_most.init("zyx._client.llm.least_to_most", "least_to_most")


class optimize_system_prompt(Loader):
    pass


optimize_system_prompt.init(
    "zyx._client.llm.optimize_system_prompt", "optimize_system_prompt"
)


class plan(Loader):
    pass


plan.init("zyx._client.llm.plan", "plan")


class self_consistency(Loader):
    pass


self_consistency.init("zyx._client.llm.self_consistency", "self_consistency")


class self_refine(Loader):
    pass


self_refine.init("zyx._client.llm.self_refine", "self_refine")


class step_back(Loader):
    pass


class tree_of_thought(Loader):
    pass


tree_of_thought.init("zyx._client.llm.tree_of_thought", "tree_of_thought")



__all__ = [
    "Agents",
    "completion",
    "embedding",
    "image",
    "llm",
    "logger",
    "Memory",
    "speak",
    "tools",
    "terminal",
    "transcribe",
    "classify",
    "code",
    "create_system_prompt",
    "extract",
    "function",
    "generate",
    "least_to_most",
    "optimize_system_prompt",
    "plan",
    "self_consistency",
    "self_refine",
    "step_back",
    "tree_of_thought",
]

from ._client.agents import Agents
from ._client.completion import completion
from litellm.main import embedding
from ._client.multimodal import image
from ._client import llm as llm
from loguru import logger
from ._client.memory import Memory
from ._client.multimodal import speak
from ._client import tools as tools
from ._client.terminal import terminal
from ._client.multimodal import transcribe
from ._client.llm import (
    classify,
    code,
    completion,
    create_system_prompt,
    extract,
    function,
    generate,
    least_to_most,
    optimize_system_prompt,
    plan,
    self_consistency,
    self_refine,
    step_back,
    tree_of_thought,
)



def prompt_cli_install(library: str) -> bool:
    """Prompts the user to install a given library"""

    import subprocess

    response = input(f"Would you like to install the {library} library? (y/n)")

    if response == "y":
        try:
            subprocess.run(["pip", "install", library])
            return True
        except Exception as e:
            raise e
    else:
        return False



__all__ = [
    "completion",
    "Client",
    "classify",
    "code",
    "create_system_prompt",
    "extract",
    "function",
    "generate",
    "least_to_most",
    "optimize_system_prompt",
    "plan",
    "self_consistency",
    "self_refine",
    "step_back",
    "tree_of_thought",
]

from ..completion import completion, CompletionClient as Client
from .classify import classify
from .code import code
from .create_system_prompt import create_system_prompt
from .extract import extract
from .function import function
from .generate import generate
from .least_to_most import least_to_most
from .optimize_system_prompt import optimize_system_prompt
from .plan import plan
from .self_consistency import self_consistency
from .self_refine import self_refine
from .step_back import step_back
from .tree_of_thought import tree_of_thought



from pydantic import BaseModel
from typing import Optional, Literal, List


def generate(
    target: BaseModel,
    instructions: Optional[str] = None,
    n: int = 1,
    model: str = "gpt-4o-mini",
    api_key: Optional[str] = None,
    base_url: Optional[str] = None,
    organization: Optional[str] = None,
    max_tokens: Optional[int] = None,
    max_retries: int = 3,
    temperature: float = 0,
    mode: Literal["json", "md_json", "tools"] = "md_json",
    verbose: bool = False,
) -> List:
    """
    Generates a list of instances of the given Pydantic model.

    Example:
        ```python
        import zyx

        class User(BaseModel):
            name: str
            age: int

        zyx.generate(User, n=5)
        ```

    Parameters:
        target (BaseModel): The Pydantic model to generate instances of.
        instructions (Optional[str]): The instructions for the generator.
        n (int): The number of instances to generate.
        model (str): The model to use for generation.
        api_key (Optional[str]): The API key to use for generation.
        base_url (Optional[str]): The base URL to use for generation.
        organization (Optional[str]): The organization to use for generation.
        max_tokens (Optional[int]): The maximum number of tokens to generate.
        max_retries (int): The maximum number of retries to make.
        temperature (float): The temperature to use for generation.
        mode (Literal["json", "md_json", "tools"]): The mode to use for generation.
        verbose (bool): Whether to print verbose output.

    Returns:
        List[BaseModel]: A list of instances of the given Pydantic model.
    """

    from ..completion import CompletionClient
    from pydantic import create_model

    ResponseModel = create_model("ResponseModel", items=(List[target], ...))

    system_message = f"""
    You are a data generator. Your task is to generate {n} valid instance(s) of the following Pydantic model:
    
    {target.model_json_schema()}
    
    Ensure that all generated instances comply with the model's schema and constraints.
    """
    user_message = (
        instructions
        if instructions
        else f"Generate {n} instance(s) of the given model."
    )

    response = CompletionClient().completion(
        messages=[
            {"role": "system", "content": system_message},
            {"role": "user", "content": user_message},
        ],
        model=model,
        api_key=api_key,
        base_url=base_url,
        organization=organization,
        max_tokens=max_tokens,
        max_retries=max_retries,
        temperature=temperature,
        verbose=verbose,
        mode="md_json" if model.startswith(("ollama/", "ollama_chat/")) else mode,
        response_model=ResponseModel,
    )
    return response.items
